/*
 *  Ray++ - Object-oriented ray tracing library
 *  Copyright (C) 1998-2001 Martin Reinecke and others.
 *  See the AUTHORS file for more information.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  See the README file for more information.
 */

/*! \file projector.cxx
    \brief Implementation of the PROJECTOR class */

/* THIS FILE is copyright (c) 1999 by Ugo Varetto - varetto@computer.org */

#include "lights/projector.h"

namespace RAYPP {

extern HANDLE<WORLD> World;

PROJECTOR::PROJECTOR () : SPOTLIGHT(),Pixels(0)
{}

PROJECTOR::PROJECTOR(const VECTOR &Loc, const VECTOR &to, const COLOUR &Col,
  const float8 &e, const float8 &c)
  : SPOTLIGHT(Loc,to,Col,e,c), Pixels(0),
    Xs(1.0), Ys(1.0), D(1.0), Y(0.0,1.0,0.0)
  {
//MR : not necessary - already done in SPOTLIGHT constructor
//  Dir=(to - Loc).Norm();
  Set_ImageAxes(Y,D);
  }

//virtual
void PROJECTOR::Cast_Light (const VECTOR &Pos, LIGHT_ARRAY &Arr) const
  {
  ci();
  VECTOR v = (Pos - Location);
  float8 d1 = v.Length();
  float8 d2 = d1*d1;
  VECTOR L = v.Norm();
  float8 cosangle = Dot(L,Dir); 
  float8 maxval= max<float8>(cosangle-CutOff,0.0);
  float8 att = pow(maxval,Exp)/
               (d2*Attenuation[2]+d1*Attenuation[1]+Attenuation[0]);
  COLOUR Col = Colour*att;

  if (Pixels)
    {
    float8 t= Num/Dot(Dir,L);
    if (t>=0.0)
      {
      VECTOR P = t*L+Location;
      P = P - O;
      float8 ucomp = Dot(P,X);
      if (abs(ucomp) <= Xs/2.0 )
        {
        float8 vcomp = Dot(P,Y);
        if (abs(vcomp) <= Ys/2.0 )
          {
          ucomp += Xs/2.0;
          vcomp += Ys/2.0;

          int4 j = Height-1-int4((vcomp/Ys)*(float8 (Height)-1.0)+.5);
          int4 i = int4((ucomp/Xs)*(float8 (Width)-1.0)+.5);
          int4 idx=3*(j*Width+i);
          COLOUR c(Pixels[idx+2],Pixels[idx+1],Pixels[idx]);
          if((c.r != Transparent.r && c.g != Transparent.g &&
          c.b != Transparent.b) || !Transparency)
            {
            c = INV_255*c;
            Col = Mix*c*Col+(1-Mix)*c;
            }
          }
        }
      }
    }

  if (!Col.TooSmall()) Arr.push_back (LIGHT_ENTRY (Col, Location, MyVolume));
  }

int PROJECTOR::Set_Image(const char *fname,int i,int j)
  {
  TGAINFOHEADER tgh;
  ifstream tga(fname,ios::binary | ios::in);
  int4 imgSize=0;
  int4 idx=0;

  if (!tga) error ("PROJECTOR: Could not open Targa file!");

  tga.read(reinterpret_cast<char *> (&tgh),sizeof(tgh));

  if (tgh.pixSize!=24 || tgh.imgType != 2)
    {
    tga.close();
    return 0;
    }

  imgSize=3*tgh.imgHeight*tgh.imgWidth;
  Pixels = new uint1[imgSize*sizeof(uint1)];
  if (!Pixels)
    {
    tga.close();
    return 0;
    }
		
  Width=tgh.imgWidth;
  Height=tgh.imgHeight;
		
  tga.read(reinterpret_cast<char *>(Pixels),imgSize*sizeof(uint1));
  tga.close();

  idx=3*(Width*j+i);
  Transparent = COLOUR(Pixels[idx+2],Pixels[idx+1],Pixels[idx]);

  return 1;
  }

void PROJECTOR::Set_ImageAxes(const VECTOR &up,float8 dd)
  {
  Y=up.Norm();
  D=dd;
  VECTOR P = dd*Dir+Location;
  O = P;
  Num = Dot(Dir,P)-Dot(Dir,Location);
  X=Cross(Y,Dir);
  }

void PROJECTOR::Set_Size(const float8 xx,const float8 yy)
  {
  Xs=xx;
  Ys=yy;
  }

void PROJECTOR::Set_Mix(const float8 t)
  {
  if (abs(t) > 1.0)
    Mix= (t/abs(t));
  else
    Mix=t;
  }

//virtual
void PROJECTOR::Transform (const TRANSFORM &trans)
  {
  cni();
  Y=trans.TransNormal(Y);
  X=trans.TransNormal(X);
  O=trans.TransPoint(O);
  Location=trans.TransPoint(Location);
  To=trans.TransPoint(To);
  }

} // namespace RAYPP
