/*
 *  Ray++ - Object-oriented ray tracing library
 *  Copyright (C) 1998-2001 Martin Reinecke and others.
 *  See the AUTHORS file for more information.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  See the README file for more information.
 */

#include "objects/bounding_slab.h"

namespace RAYPP {

class INTERSECT_INFO;

//virtual
void BOUNDING_SLAB::Init ()
  {
  if (initialized) return;
  Bounder.Reset();
  Contains_Volume = false;
  V_in_BBox = true;
  for (vector<HANDLE<OBJECT> >::iterator i=Object.begin(); i<Object.end(); ++i)
    {
    (*i)->Init ();
    Bounder.Include ((*i)->BBox());
    if ((*i)->Has_Volume())
      {
      Contains_Volume = true;
      if (!(*i)->Volume_in_BBox()) V_in_BBox = false;
      }
    }
  initialized = true;
  }

//virtual
void BOUNDING_SLAB::Deinit ()
  {
  if (!initialized) return;
  for (vector<HANDLE<OBJECT> >::iterator i=Object.begin(); i<Object.end(); ++i)
    (*i)->Deinit ();
  initialized = false;
  }

//virtual
void BOUNDING_SLAB::Transform (const TRANSFORM &trans)
  {
  cni();
  for (vector<HANDLE<OBJECT> >::iterator i=Object.begin(); i<Object.end(); ++i)
    (*i)->Transform (trans);
  }

//virtual
bool BOUNDING_SLAB::Inside_Volume (const VECTOR &Loc,
  const INSIDE_INFO &nowInside, INSIDE_INFO &result) const
  {
  ci();
  if (!Contains_Volume) return false;
  if (V_in_BBox)
    if (!Bounder.Inside (Loc)) return false;

  INSIDE_INFO tmpins = nowInside;

  bool found = false;

  for (vector<HANDLE<OBJECT> >::const_iterator i=Object.begin(); i<Object.end(); ++i)
    {
    if ((*i)->Inside_Volume (Loc, tmpins, result))
      {
      tmpins = result;
      found = true;
      }
    }
  return found;
  }

//virtual
bool BOUNDING_SLAB::Has_Volume () const
  {
  ci();
  return Contains_Volume;
  }
//virtual
bool BOUNDING_SLAB::Volume_in_BBox () const
  {
  ci();
  return V_in_BBox;
  }

//virtual
bool BOUNDING_SLAB::Test (RAY &Ray, float8 &dist) const
  {
  ci();
  return Bounder.Ray_in_Bounds (Ray, dist);
  }

//virtual
bool BOUNDING_SLAB::Intersect (RAY &Ray,
  OBJECT_QUEUE &Queue, INTERSECT_INFO &) const
  {
  ci();

  float8 dist;
  for (vector<HANDLE<OBJECT> >::const_iterator i=Object.begin(); i<Object.end(); ++i)
    {
    if ((*i)->Test (Ray, dist))
      Queue.push (oqentry (dist, *i));
    }
  return false;
  }

void BOUNDING_SLAB::Add (const HANDLE<OBJECT> &new_obj)
  {
  cni();
  Object.push_back (new_obj);
  }

} // namespace RAYPP
