/*
 *  Ray++ - Object-oriented ray tracing library
 *  Copyright (C) 1998-2001 Martin Reinecke and others.
 *  See the AUTHORS file for more information.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  See the README file for more information.
 */

/*
    AUTHOR:
        Dair Grant <mailto:dair@webthing.net>
*/

#include "outputs/mem_output.h"

namespace RAYPP {

MEM_OUTPUT::MEM_OUTPUT()
 : mWidth(0), mHeight(0), mFilter(0), mFilterRefCon(0), mBaseAddr(0),
   mRowBytes(0), mPixelBytes(0), mShiftRed(0), mShiftGreen(0), mShiftBlue(0),
   mMaskRed(0), mMaskGreen(0), mMaskBlue(0)
  {
  UpdateScale();
  }

MEM_OUTPUT::MEM_OUTPUT( uint4 theWidth,   uint4 theHeight,
                        OutputFilterFunc        theFilter,
                        uint4                   filterRefCon,
                        void                    *baseAddr,
                        uint4 rowBytes,   uint4 pixelBytes,
                        uint4 shiftRed,   uint4 maskRed,
                        uint4 shiftGreen, uint4 maskGreen,
                        uint4 shiftBlue,  uint4 maskBlue)
 : mWidth(theWidth), mHeight(theHeight), mFilter(theFilter),
   mFilterRefCon(filterRefCon), mBaseAddr(baseAddr),
   mRowBytes(rowBytes), mPixelBytes(pixelBytes),
   mShiftRed(shiftRed), mShiftGreen(shiftGreen), mShiftBlue(shiftBlue),
   mMaskRed(maskRed), mMaskGreen(maskGreen), mMaskBlue(maskBlue)
  {
  UpdateScale();
  }

//virtual
MEM_OUTPUT::~MEM_OUTPUT()
  {}


void MEM_OUTPUT::Set_Resolution(uint4 theWidth, uint4 theHeight)
  {
  cni();

  mWidth  = theWidth;
  mHeight = theHeight;
  }

void MEM_OUTPUT::Set_Filter(OutputFilterFunc theFilter, uint4 filterRefCon)
  {
  cni();

  mFilter       = theFilter;
  mFilterRefCon = filterRefCon;
  }

void MEM_OUTPUT::Set_Image(void *baseAddr, uint4 rowBytes, uint4 pixelBytes)
  {
  cni();

  mBaseAddr   = baseAddr;
  mRowBytes   = rowBytes;
  mPixelBytes = pixelBytes;
  }

void MEM_OUTPUT::Set_RedFormat(uint4 shiftRed, uint4 maskRed)
  {
  cni();

  // Set our red format and update our scale
  mShiftRed = shiftRed;
  mMaskRed  = maskRed;

  UpdateScale();
  }

void MEM_OUTPUT::Set_GreenFormat(uint4 shiftGreen, uint4 maskGreen)
  {
  cni();

  // Set our green format and update our scale
  mShiftGreen = shiftGreen;
  mMaskGreen  = maskGreen;

  UpdateScale();
  }

void MEM_OUTPUT::Set_BlueFormat(uint4 shiftBlue, uint4 maskBlue)
  {
  cni();

  // Set our blue format and update our scale
  mShiftBlue = shiftBlue;
  mMaskBlue  = maskBlue;

  UpdateScale();
  }

void MEM_OUTPUT::Render() const
  {
  bool        areDone;
  COLOUR      *theScan;
  float8      du, dv;
  uint4       x, y;

  ci();

  if (mBaseAddr == 0)
    return;

  // Prepare to render the image
  theScan = new COLOUR[mWidth];
  if (theScan == 0)
    return;

  areDone = false;
  du      = float8(1.0) / mWidth;
  dv      = float8(1.0) / mHeight;

  // Render the image
  for (y = 0; y < mHeight && !areDone; y++)
    {
    // Render the scanline
    for (x = 0; x < mWidth; x++)
      {
      theScan[x] = Renderer->Get_Pixel(x * du, y * dv, du, dv);
      theScan[x].Clip();
      }

    // Pass the scanline to our filter, if any
    if (mFilter != 0)
      areDone = mFilter(mFilterRefCon, y, mWidth, theScan);

    // Rasterise the scanline
    DrawScan(y, theScan);
    }

  // Clean up
  delete [] theScan;
  } 

//=============================================================================
//      MEM_OUTPUT::UpdateScale : Update our component scaling factors.
//-----------------------------------------------------------------------------
//      Note :  We assume that the mask for each component is a continguous
//              bitpattern, and so we can determine the maximum value of each
//              component by shifting the component mask down.
//-----------------------------------------------------------------------------
void MEM_OUTPUT::UpdateScale()
  {
  // Calculate the scaling factors we need
  mScaleRed   = mMaskRed;
  mScaleGreen = mMaskGreen;
  mScaleBlue  = mMaskBlue;

  while (mScaleRed != 0 && ((mScaleRed & 0x01) == 0))
    mScaleRed >>= 1;

  while (mScaleGreen != 0 && ((mScaleGreen & 0x01) == 0))
    mScaleGreen >>= 1;

  while (mScaleBlue != 0 && ((mScaleBlue & 0x01) == 0))
    mScaleBlue >>= 1;
  }

//=============================================================================
//      MEM_OUTPUT::DrawScan : Draw a scanline into the destination image.
//-----------------------------------------------------------------------------
void MEM_OUTPUT::DrawScan(uint4 scanY, COLOUR *theScan) const
  {
  uint4       redBits, greenBits, blueBits;
  uint4       x, thePixel;
  byte        *dstPixel;

  // Prepare to draw the scanline
  dstPixel = (static_cast<byte *> (mBaseAddr)) + (scanY * mRowBytes);

  // Draw the scanline a pixel at a time
  for (x = 0; x < mWidth; x++)
    {
    // Get the components for the pixel
    redBits   = uint4 (theScan[x].r * mScaleRed);
    greenBits = uint4 (theScan[x].g * mScaleGreen);
    blueBits  = uint4 (theScan[x].b * mScaleBlue);

    // Build the pixel
    redBits   = (redBits   << mShiftRed)   & mMaskRed;
    greenBits = (greenBits << mShiftGreen) & mMaskGreen;
    blueBits  = (blueBits  << mShiftBlue)  & mMaskBlue;
    thePixel  = (redBits | greenBits | blueBits);

    // Write it out
    if (mPixelBytes == 2)
//      *((TQ3Uns16 *) dstPixel) = thePixel;
      *(reinterpret_cast<uint2 *> (dstPixel)) = thePixel;
    else
//      *((TQ3Uns32 *) dstPixel) = thePixel;
      *(reinterpret_cast<uint4 *> (dstPixel)) = thePixel;

    // Move on to the next pixel
    dstPixel += mPixelBytes;
    }
  }

} // namespace RAYPP
