/*
 *  Ray++ - Object-oriented ray tracing library
 *  Copyright (C) 1998-2001 Martin Reinecke and others.
 *  See the AUTHORS file for more information.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  See the README file for more information.
 */

#include "kernel/globals.h"
#include "kernel/handle.h"
#include "worlds/scene.h"
#include "renderers/raytracer.h"
#include "outputs/tga_output.h"
#include "shapes/implicit.h"
#include "shapes/plane.h"
#include "surfaces/phong.h"
#include "objects/simple_object.h"
#include "lights/pointlight.h"
#include "cameras/std_camera.h"

namespace RAYPP {

class Dens1: public IMPLICIT::DENSFUNC
  {
  public:
    float8 Density (const VECTOR &Loc) const
      {
      float8 rad=sqrt(Loc.x*Loc.x + Loc.z*Loc.z);
      return Loc.y - cos(3*rad)/(rad+1);
      }
    VECTOR Maxgrad () const
      {
      return VECTOR(3,1,3);
      }
    VECTOR Mingrad () const
      {
      return VECTOR(-3,1,-3);
      }
  };

} // namespace RAYPP

int main ()
  {
  using namespace RAYPP;

  HANDLE<OUTPUT> Output;

  Output = Cnew TGA_OUTPUT (640, 480, "imptest3.tga");
  Output->Init();

  RAYTRACER *Raytracer = Cnew RAYTRACER;
  SCENE *Scn = Cnew SCENE;

  Raytracer->Add (Cnew STD_CAMERA (VECTOR (3,4,-6.5), VECTOR (0,0,0)));

  SIMPLE_OBJECT *obj;
  uint4 Priority = 1;

  IMPLICIT *imp=Cnew IMPLICIT;
  imp->Set_Minmax(AXISBOX(VECTOR(-50,-1.01,-50),VECTOR(50,1.01,50)));
  imp->Set_Densfunc(Cnew Dens1);

  obj = Cnew SIMPLE_OBJECT;
  obj->Set_Shape (imp);
  obj->Set_Surface (Cnew PHONG (0.2, 0.8, 0.6,0,0,50, COLOUR(1.0,0.5,0.8)));
  obj->Set_Priority (Priority);
  Scn->Add (obj);

  Scn->Add (Cnew POINTLIGHT (VECTOR (0,6,-10), COLOUR (80,80,80)));
  Scn->Add (Cnew POINTLIGHT (VECTOR (3,10,5), COLOUR (80,80,80)));

  World = Scn;
  World->Init();

  Renderer = Raytracer;
  Renderer->Init();

  Output->Render();
  }

