/*
 *  Ray++ - Object-oriented ray tracing library
 *  Copyright (C) 1998-2001 Martin Reinecke and others.
 *  See the AUTHORS file for more information.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  See the README file for more information.
 */

#include "shapes/implicit.h"

namespace RAYPP {

const float8 IMPLICIT::Epsilon = Small_dist;

inline VECTOR IMPLICIT::Get_Normal (const VECTOR &rawnorm) const
  {
  VECTOR tmp = Trans.TransNormal (rawnorm);
  tmp.Normalize();
  if (Inverted) tmp.Flip();
  return tmp;
  }

float8 IMPLICIT::Get_Deriv (const VECTOR &dir, float8 density) const
  {
  float8 deriv;
  VECTOR ming = Dens->Mingrad();
  VECTOR maxg = Dens->Maxgrad();
  if (density > 0)
    {
    deriv  = ((dir.x > 0) ? (-ming.x*dir.x) : (-maxg.x*dir.x))
           + ((dir.y > 0) ? (-ming.y*dir.y) : (-maxg.y*dir.y))
           + ((dir.z > 0) ? (-ming.z*dir.z) : (-maxg.z*dir.z));
    }
  else
    {
    deriv  = ((dir.x > 0) ? ( maxg.x*dir.x) : ( ming.x*dir.x))
           + ((dir.y > 0) ? ( maxg.y*dir.y) : ( ming.y*dir.y))
           + ((dir.z > 0) ? ( maxg.z*dir.z) : ( ming.z*dir.z));
    }
  return max (deriv, Epsilon);
  }

IMPLICIT::IMPLICIT ()
  : Minmax (VECTOR (0,0,0), VECTOR (1,1,1)), stepwidth (0.01), Inverted (false),
    Solid (false)
  {}

//virtual
void IMPLICIT::Init ()
  {
  if (initialized) return;

  if (!Dens)
    error ("IMPLICIT: No density function given");

  initialized = true;
  }

//virtual
void IMPLICIT::Transform (const TRANSFORM &trans)
  {
  cni();
  Trans.Add_Transform (trans);
  }

//virtual
AXISBOX IMPLICIT::BBox () const
  {
  ci();
  AXISBOX Mybox=Minmax;
  Mybox.Transform (Trans);
  return Mybox;
  }

//virtual
bool IMPLICIT::Test (const GEOM_RAY &Ray, float8 &dist, bool &realhit) const
  {
  ci();
  realhit = false;
  return (Minmax.Ray_in_Bounds (Trans.InvTransRay (Ray), dist));
  }

//virtual
bool IMPLICIT::Inside (const VECTOR &Loc) const
  {
  ci();

  if (!Solid) return false;

  VECTOR tmpvec = Trans.InvTransPoint (Loc);
  if (!Minmax.Inside (tmpvec)) return Inverted;
  return ((Dens->Density (tmpvec) > 0) ^ Inverted);
  }

//virtual
bool IMPLICIT::Intersect 
  (const GEOM_RAY &Ray, float8 &dist, VECTOR &Normal) const
  {
  ci();

  GEOM_RAY Local_Ray = Trans.InvTransRay (Ray);
  if (!Minmax.Clip_Ray (Local_Ray)) return false;

  VECTOR newpoint = Local_Ray.evalmin();
  float8 new_dist = Local_Ray.mindist;
  float8 new_dens = Dens->Density (newpoint);

  if (Solid)
    {
    if (new_dist!=Ray.mindist)
      {
      if (new_dens>0)
        {
        dist = new_dist;
        Normal = Get_Normal (Minmax.Normal (newpoint));
        return true;
        }
      }
    }

  float8 maxgrad  = Get_Deriv (Local_Ray.dir, new_dens);

  while (true)
    {
    VECTOR oldpoint = newpoint;
    float8 old_dist = new_dist;
    float8 old_dens = new_dens;

    float8 advance = max (abs(old_dens)/maxgrad, stepwidth);
    new_dist = old_dist + advance;
    if (new_dist > Local_Ray.maxdist)
      break;

    newpoint = Local_Ray.eval(new_dist);

    new_dens = Dens->Density (newpoint);

    if ((new_dens*old_dens) < 0)
      {
      while ((new_dist - old_dist) > Epsilon)
        {
        float8 temp_dist = (old_dist + new_dist)/2;
        VECTOR temppoint = Local_Ray.eval(temp_dist);
        float8 temp_dens = Dens->Density(temppoint);
        if ((old_dens*temp_dens) < 0)
          {
          new_dist = temp_dist;
          newpoint = temppoint;
          new_dens = temp_dens;
          }
        else
          {
          old_dist = temp_dist;
          oldpoint = temppoint;
          old_dens = temp_dens;
          }
        }
      dist = old_dist;

      Normal = Get_Normal (Dens->Normal (oldpoint, old_dens));
      return true;
      }
    }

  if (Solid)
    {
    if (Local_Ray.maxdist!=Ray.maxdist)
      {
      if (new_dens>0)
        {
        dist = Local_Ray.maxdist;
        Normal = Get_Normal (Minmax.Normal (Local_Ray.evalmax()));
        return true;
        }
      }
    }

  return false;
  }

//virtual
void IMPLICIT::All_Intersections (const GEOM_RAY &Ray, vector<INTER> &Inter)
  const
  {
  ci();

  GEOM_RAY Local_Ray = Trans.InvTransRay (Ray);
  if (!Minmax.Clip_Ray (Local_Ray)) return;

  VECTOR newpoint = Local_Ray.evalmin();
  float8 new_dist = Local_Ray.mindist;
  float8 new_dens = Dens->Density (newpoint);

  if (Solid)
    {
    if (new_dist != Ray.mindist)
      {
      if (new_dens>0)
        Inter.push_back
          (INTER (new_dist, Get_Normal (Minmax.Normal(newpoint))));
      }
    }

  float8 maxgrad  = Get_Deriv (Local_Ray.dir, new_dens);

  while (true)
    {
    VECTOR oldpoint = newpoint;
    float8 old_dist = new_dist;
    float8 old_dens = new_dens;

    float8 advance = max (abs(old_dens)/maxgrad,stepwidth);
    new_dist = old_dist + advance;
    if (new_dist > Local_Ray.maxdist) break;

    newpoint = Local_Ray.eval(new_dist);

    new_dens = Dens->Density (newpoint);

    if ((new_dens*old_dens) < 0)
      {
      while ((new_dist - old_dist) > Epsilon)
        {
        float8 temp_dist = (old_dist + new_dist)/2;
        VECTOR temppoint = Local_Ray.eval(temp_dist);
        float8 temp_dens = Dens->Density(temppoint);
        if ((old_dens*temp_dens) < 0)
          {
          new_dist = temp_dist;
          newpoint = temppoint;
          new_dens = temp_dens;
          }
        else
          {
          old_dist = temp_dist;
          oldpoint = temppoint;
          old_dens = temp_dens;
          }
        }

      Inter.push_back
        (INTER (old_dist, Get_Normal (Dens->Normal (oldpoint, old_dens))));
      maxgrad  = Get_Deriv (Local_Ray.dir, new_dens);
      }
    }

  if (Solid)
    {
    if (Local_Ray.maxdist!=Ray.maxdist)
      {
      if (new_dens>0)
        Inter.push_back (INTER (Local_Ray.maxdist,
          Get_Normal (Minmax.Normal (Local_Ray.evalmax()))));
      }
    }
  }

void IMPLICIT::Set_Densfunc (const HANDLE<DENSFUNC> &newdens)
  {
  cni();
  Dens = newdens;
  }

void IMPLICIT::Set_Minmax (const AXISBOX &newbox)
  {
  cni();
  Minmax = newbox;
  }

void IMPLICIT::Set_Stepwidth (float8 st)
  {
  cni();
  stepwidth = st;
  }

} // namespace RAYPP
