/*
 *  Ray++ - Object-oriented ray tracing library
 *  Copyright (C) 1998-2001 Martin Reinecke and others.
 *  See the AUTHORS file for more information.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  See the README file for more information.
 */

#include "shapes/parametric.h"

namespace RAYPP {

PARAMETRIC::PARAMETRIC ()
  : du(-1), dv(-1) {}

PARAMETRIC::PARAMETRIC (const HANDLE<PARFUNC> &func)
  : Func(func), du(-1), dv(-1) {}

//virtual
void PARAMETRIC::Init ()
  {
  if (initialized) return;

  if (!Func) error ("PARAMETRIC: no PARFUNC given!");

  IV8 iu, iv;
  Func->Get_Ranges (iu,iv);
  if (du<0) du = iu.len() / 500;
  if (dv<0) dv = iv.len() / 500;

  Minmax = Func->Get_Extent (iu,iv);

  float8 lu = iu.len(), lv = iv.len();

  dircache.clear();
  maxlevel=0;

  while ((lu>du) || (lv>dv))
    {
    if ((lu>du) && ((lu>lv) || (lv<=dv)))
      {
      dircache.push_back (u_split);
      lu /= 2;
      ++maxlevel;
      }
    else if ((lv>dv) && ((lv>lu) || (lu<=du)))
      {
      dircache.push_back (v_split);
      lv /= 2;
      ++maxlevel;
      }
    }

  initialized = true;
  }

//virtual
void PARAMETRIC::Transform (const TRANSFORM &trans)
  {
  cni();
  Trans.Add_Transform (trans);
  }

//virtual
AXISBOX PARAMETRIC::BBox () const
  {
  ci();

  AXISBOX Mybox = Minmax;
  Mybox.Transform (Trans);
  return Mybox;
  }

//virtual
bool PARAMETRIC::Test (const GEOM_RAY &Ray, float8 &dist, bool &realhit) const
  {
  ci();
  realhit = false;
  return (Minmax.Ray_in_Bounds (Trans.InvTransRay (Ray), dist));
  }

bool PARAMETRIC::Check_Interval (priority_queue<entry> &Queue,
  const IV8 &iu, const IV8 &iv, int4 lvl,
  const GEOM_RAY &Local_Ray, float8 &dist, VECTOR &Normal) const
  {
  float8 mind, maxd;
  AXISBOX Box = Func->Get_Extent (iu, iv);
  if (!Box.Ray_in_Bounds (Local_Ray, mind, maxd)) return false;

  if (lvl>=maxlevel)
    {
    float8 ru, rv;
    if (Intersect_Quadrangle(Func->Get_Point (iu.lo, iv.lo),
                             Func->Get_Point (iu.hi, iv.lo),
                             Func->Get_Point (iu.lo, iv.hi),
                             Func->Get_Point (iu.hi, iv.hi),
                             Local_Ray, dist, ru, rv))
      {
      ru = iu.lo + ru*iu.len();
      rv = iv.lo + rv*iv.len();
      VECTOR Loc = Func->Get_Point (ru, rv);
      Normal = Cross (Func->Get_Point (ru+Small_dist, rv) - Loc,
                      Func->Get_Point (ru, rv+Small_dist) - Loc);
      Normal = (Trans.TransNormal(Normal)).Norm();
      return true;
      }
    return false;
    }
  Queue.push (entry (mind, iu, iv, lvl));
  return false;
  }

//virtual
bool PARAMETRIC::Intersect (const GEOM_RAY &Ray, float8 &dist,
  VECTOR &Normal) const
  {
  ci();

  GEOM_RAY Local_Ray = Trans.InvTransRay (Ray);
  if (!Minmax.Clip_Ray (Local_Ray)) return false;

  priority_queue<entry> Queue;
  IV8 iu, iv;
  Func->Get_Ranges (iu, iv);
  bool found = false;

  if (Check_Interval (Queue, iu, iv, 0, Local_Ray, dist, Normal))
    {
    Local_Ray.maxdist = dist;
    found = true;
    }

  while (!Queue.empty())
    {
    iu = Queue.top().iu;
    iv = Queue.top().iv;
    int4 lvl = Queue.top().lvl;
    Queue.pop();

    IV8 i1, i2, i3, i4;

    switch (dircache[lvl])
      {
      case u_split:
        iu.split (i1, i3);
        i2 = i4 = iv;
        break;
      case v_split:
        i1 = i3 = iu;
        iv.split (i2, i4);
        break;
      }

    ++lvl;

    if (Check_Interval (Queue, i1, i2, lvl, Local_Ray, dist, Normal))
      {
      Local_Ray.maxdist = dist;
      found = true;
      }
    if (Check_Interval (Queue, i3, i4, lvl, Local_Ray, dist, Normal))
      {
      Local_Ray.maxdist = dist;
      found = true;
      }
    }
  return found;
  }

//virtual
void PARAMETRIC::All_Intersections (const GEOM_RAY &Ray,
  vector<INTER> &Inter) const
  {
  ci();

  GEOM_RAY Local_Ray = Trans.InvTransRay (Ray);
  if (!Minmax.Clip_Ray (Local_Ray)) return;

  priority_queue<entry> Queue;

  IV8 iu, iv;
  Func->Get_Ranges(iu, iv);

  float8 dist;
  VECTOR Normal;

  if (Check_Interval (Queue, iu, iv, 0, Local_Ray, dist, Normal))
    Inter.push_back (INTER (dist, Normal));

  while (!Queue.empty())
    {
    iu = Queue.top().iu;
    iv = Queue.top().iv;
    int4 lvl = Queue.top().lvl;
    Queue.pop();

    IV8 i1, i2, i3, i4;

    switch (dircache[lvl])
      {
      case u_split:
        iu.split (i1, i3);
        i2 = i4 = iv;
        break;
      case v_split:
        i1 = i3 = iu;
        iv.split (i2, i4);
        break;
      }

    ++lvl;

    if (Check_Interval (Queue, i1, i2, lvl, Local_Ray, dist, Normal))
      Inter.push_back (INTER (dist, Normal));
    if (Check_Interval (Queue, i3, i4, lvl, Local_Ray, dist, Normal))
      Inter.push_back (INTER (dist, Normal));
    }
  }

void PARAMETRIC::Set_Func (const HANDLE<PARFUNC> &newfunc)
  {
  cni();
  Func = newfunc;
  }

void PARAMETRIC::Set_Resolution (float8 Du, float8 Dv)
  {
  cni();
  du = Du; dv = Dv;
  }

} // namespace RAYPP
