/*
 *  Ray++ - Object-oriented ray tracing library
 *  Copyright (C) 1998-2001 Martin Reinecke and others.
 *  See the AUTHORS file for more information.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  See the README file for more information.
 */

#include "shapes/sphere.h"

namespace RAYPP {

inline VECTOR SPHERE::Get_Normal (const VECTOR &rawnorm) const
  {
  VECTOR tmp = Trans.TransNormal (rawnorm);
  tmp.Normalize();
  if (Inverted) tmp.Flip();
  return tmp;
  }

SPHERE::SPHERE (const VECTOR &center, float8 radius)
  {
  Trans = Scaling_Transform (VECTOR (radius, radius, radius));
  Trans.Add_Transform (Translation_Transform (center));
  }

//virtual
void SPHERE::Init ()
  {
  initialized = true;
  }

//virtual
void SPHERE::Transform (const TRANSFORM &trans)
  {
  cni();
  Trans.Add_Transform (trans);
  }

//virtual
AXISBOX SPHERE::BBox () const
  {
  ci();

  AXISBOX Mybox;
  TRANSFORM t (Trans);
  Mybox.Include(t.TransPoint((t.InvTransNormal(VECTOR( 1, 0, 0))).Norm()));
  Mybox.Include(t.TransPoint((t.InvTransNormal(VECTOR( 0, 1, 0))).Norm()));
  Mybox.Include(t.TransPoint((t.InvTransNormal(VECTOR( 0, 0, 1))).Norm()));
  Mybox.Include(t.TransPoint((t.InvTransNormal(VECTOR(-1, 0, 0))).Norm()));
  Mybox.Include(t.TransPoint((t.InvTransNormal(VECTOR( 0,-1, 0))).Norm()));
  Mybox.Include(t.TransPoint((t.InvTransNormal(VECTOR( 0, 0,-1))).Norm()));
  return Mybox;
  }

//virtual
bool SPHERE::Test (const GEOM_RAY &Ray, float8 &dist, bool &realhit) const
  {
  ci();

  VECTOR s = Trans.InvTransPoint (Ray.start),
         d = Trans.InvTransDirection (Ray.dir);
  float8 s_quad = Dot (s, s),
         d_quad = Dot (d, d),
         d_s    = Dot (s, d);
  float8 discrim = 4.0 * (d_s*d_s - d_quad*(s_quad - 1.0));

  if (discrim <= Small_float8) return false;

  discrim = sqrt (discrim);

  if ((dist = (-2.0*d_s - discrim)/(2.0*d_quad)) > Ray.maxdist)
    return false;

  if (dist < Ray.mindist)
    {
    if ((dist = (-2.0*d_s + discrim)/(2.0*d_quad)) > Ray.maxdist)
      return false;
    if (dist < Ray.mindist) return false;
    }

  realhit = true;
  return true;
  }

//virtual
bool SPHERE::Inside (const VECTOR &Loc) const
  {
  ci();

  VECTOR tmpvec = Trans.InvTransPoint (Loc);
  return ((Dot (tmpvec, tmpvec) <= 1.0) ^ Inverted);
  }

//virtual
bool SPHERE::Intersect (const GEOM_RAY &Ray, float8 &dist,
  VECTOR &Normal) const
  {
  ci();

  VECTOR s = Trans.InvTransPoint (Ray.start),
         d = Trans.InvTransDirection (Ray.dir);
  float8 s_quad = Dot (s,s),
         d_quad = Dot (d,d),
         d_s    = Dot (s,d);
  float8 discrim = 4.0 * (d_s*d_s - d_quad*(s_quad - 1.0));

  if (discrim <= Small_float8) return false;

  discrim = sqrt (discrim);

  if ((dist = (-2.0*d_s - discrim)/(2.0*d_quad)) > Ray.maxdist) return false;

  if (dist < Ray.mindist)
    {
    if ((dist = (-2.0*d_s + discrim)/(2.0*d_quad)) > Ray.maxdist) return false;
    if (dist < Ray.mindist) return false;
    }

  Normal = Get_Normal (s + dist*d);
  return true;
  }

//virtual
void SPHERE::All_Intersections (const GEOM_RAY &Ray, vector<INTER> &Inter) const
  {
  ci();

  VECTOR s = Trans.InvTransPoint (Ray.start),
         d = Trans.InvTransDirection (Ray.dir);
  float8 s_quad = Dot (s,s),
         d_quad = Dot (d,d),
         d_s = Dot (s,d);
  float8 discrim = 4 * (d_s*d_s - d_quad*(s_quad - 1));
  float8 depth;

  if (discrim <= Small_float8) return;

  discrim = sqrt (discrim);

  if ((depth = (-2.0*d_s - discrim)/(2.0*d_quad)) > Ray.maxdist) return;
  if (depth > Ray.mindist)
    Inter.push_back (INTER (depth, (Get_Normal (s + depth*d))));
  if ((depth = (-2.0*d_s + discrim)/(2.0*d_quad)) > Ray.maxdist) return;
  if (depth > Ray.mindist)
    Inter.push_back (INTER (depth, (Get_Normal (s + depth*d))));
  }

} // namespace RAYPP
