/*
 *  Ray++ - Object-oriented ray tracing library
 *  Copyright (C) 1998-2001 Martin Reinecke and others.
 *  See the AUTHORS file for more information.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  See the README file for more information.
 */

#include "surfaces/matte.h"
#include "surfaces/pigments/pure_colour.h"

namespace RAYPP {

MATTE::MATTE () : Ka (float4(0.8)), Kd (float4(0.8)) {}

MATTE::MATTE (float4 a, float4 d, const HANDLE<PIGMENT> &pig)
  : Ka (a), Kd (d), Pigment (pig) {}

MATTE::MATTE (float4 a, float4 d, const COLOUR &Col)
  : Ka (a), Kd (d), Pigment (Cnew PURE_COLOUR(Col)) {}

//virtual
void MATTE::Init ()
  {
  if (initialized) return;
  if (!Pigment) error ("MATTE: no pigment given");
  initialized = true;
  }

//virtual 
void MATTE::Transform (const TRANSFORM &trans)
  {
  cni();
  if (Pigment) Pigment->Transform (trans);
  }

//virtual
void MATTE::Get_Full_Shading_Info (const SHADING_INFO &In,
  FULL_SHADING_INFO &Out) const
  {
  SHADING_INFO &tmp = Out;
  tmp = In;

  Out.MC_diffuse = false;
  Out.reflect    = false;
  Out.MC_reflect = false;
  Out.refract    = false;
  Out.MC_refract = false;

  Out.Bumped_Normal = In.Normal; 
  Out.Reflected = VECTOR (0,0,0); 
  Out.Refracted = VECTOR (0,0,0); 
  }

//virtual
COLOUR MATTE::Get_Total_Importance (const FULL_SHADING_INFO &Info,
  const VECTOR &Dir) const
  {
  ci();
  float8 dotprod = Dot (Dir, Info.Normal);
  if (dotprod<0) return COLOUR(0,0,0);
  return Kd*Info.Importance*dotprod*Pigment->Get_Colour (Info);
  }

//virtual
VECTOR MATTE::Get_MC_Reflected_Dir (const FULL_SHADING_INFO &) const
  { ci(); return VECTOR (0,0,0); }

//virtual
VECTOR MATTE::Get_MC_Refracted_Dir (const FULL_SHADING_INFO &) const
  { ci(); return VECTOR (0,0,0); }

//virtual
VECTOR MATTE::Get_MC_Diffuse_Dir (const FULL_SHADING_INFO &) const
  { ci(); return VECTOR (0,0,0); }

//virtual
COLOUR MATTE::Get_Emitted_Light (const FULL_SHADING_INFO &Info,
  const INCIDENT_ARRAY &Arr, const COLOUR &, const COLOUR &,
  const INCIDENT_ARRAY &, const INCIDENT_ARRAY &, const INCIDENT_ARRAY &) const
  {
  ci();
  float8 dotprod;
  COLOUR newcol = Pigment->Get_Colour (Info);
  COLOUR Col = Ka*newcol*Arr.Ambient;

  for (INCIDENT_ARRAY::const_iterator i=Arr.begin(); i<Arr.end(); ++i)
    {
    if ((dotprod = -Dot((i->Direction), Info.Normal)) >= 0.0)
      Col += newcol*Kd*(i->Intensity)*dotprod;
    }

  return Col;
  }

//virtual
COLOUR MATTE::Get_Transmitted_Light
  (const SHADING_INFO &, const COLOUR &) const
  {
  ci();
  return COLOUR(0,0,0);
  }

void MATTE::Set_Ka (float4 new_Ka)
  {
  cni();
  Ka = new_Ka;
  }

void MATTE::Set_Kd (float4 new_Kd)
  {
  cni();
  Kd = new_Kd;
  }

void MATTE::Set_Pigment (const HANDLE<PIGMENT> &ThePig)
  {
  cni();
  Pigment = ThePig;
  }

} // namespace RAYPP
