/*
 *  Ray++ - Object-oriented ray tracing library
 *  Copyright (C) 1998-2001 Martin Reinecke and others.
 *  See the AUTHORS file for more information.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  See the README file for more information.
 */

#include "surfaces/pigments/std_pigments.h"

namespace RAYPP {

//virtual
COLOUR AGATE::Get_Colour (const SHADING_INFO &Info) const
  {
  VECTOR tmpvec = Trans.InvTransPoint (Info.Intersect_Point);
  float8 val, tmpturb = float8(0.0);

  if (Turb != 0)
    tmpturb = Noise.fBm (tmpvec, Lacunarity, Omega, Octaves)*Turb;

  val = clamp (0.5*(cycloidal (1.3*tmpturb + 1.1*tmpvec.z) + 1.0), 0, 1);

  return CMap.Get_Colour (val);
  }


//virtual
COLOUR BOZO::Get_Colour (const SHADING_INFO &Info) const
  {
  VECTOR tmpvec = Trans.InvTransPoint (Info.Intersect_Point);

  if (Turb != 0)
    tmpvec += Noise.DfBm (tmpvec, Lacunarity, Omega, Octaves)*Turb;

  return CMap.Get_Colour (clamp(0.5+Noise.Noise (tmpvec),0,1));
  }


//virtual
COLOUR GRANITE::Get_Colour (const SHADING_INFO &Info) const
  {
  VECTOR tmpvec = Trans.InvTransPoint (Info.Intersect_Point);
  float8 noise;

  noise = 1.4*Noise.Turbulence (4*tmpvec, Lacunarity, Omega, Octaves);

  return CMap.Get_Colour (clamp (noise, 0, 1));
  }


//virtual
COLOUR MARBLE::Get_Colour (const SHADING_INFO &Info) const
  {
  VECTOR tmpvec = Trans.InvTransPoint (Info.Intersect_Point);
  float8 val;

  if (Turb != 0)
    val = trianglewave 
      (tmpvec.x + Noise.fBm (tmpvec, Lacunarity, Omega, Octaves)*Turb);
  else val = trianglewave (tmpvec.x);

  return CMap.Get_Colour (val);
  }


//virtual
COLOUR ONION::Get_Colour (const SHADING_INFO &Info) const
  {
  VECTOR tmpvec = Trans.InvTransPoint (Info.Intersect_Point);
  float8 rad = tmpvec.Length();

  if (Turb != 0)
    rad += Noise.fBm (tmpvec, Lacunarity, Omega, Octaves)*Turb;

  return CMap.Get_Colour (trianglewave (rad));
  }


//virtual
COLOUR WOOD::Get_Colour (const SHADING_INFO &Info) const
  {
  VECTOR tmpvec = Trans.InvTransPoint (Info.Intersect_Point);
  float8 rad = sqrt (tmpvec.x*tmpvec.x + tmpvec.z*tmpvec.z);

  if (Turb != 0)
    rad += Noise.fBm (tmpvec, Lacunarity, Omega, Octaves)*Turb;

  return CMap.Get_Colour (trianglewave (rad));
  }


//virtual
COLOUR CHECKER::Get_Colour (const SHADING_INFO &Info) const
  {
  VECTOR tmpvec = Trans.InvTransPoint (Info.Intersect_Point);

  int4 res = int4 (floor(2*tmpvec.x + Small_dist) +
                   floor(2*tmpvec.y + Small_dist) +
                   floor(2*tmpvec.z + Small_dist));

  if (res % 2) return CMap.Get_Colour (0.0);
  return CMap.Get_Colour (1.0);
  }

} // namespace RAYPP
