/*
 *  Ray++ - Object-oriented ray tracing library
 *  Copyright (C) 1998-2001 Martin Reinecke and others.
 *  See the AUTHORS file for more information.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  See the README file for more information.
 */

#include "worlds/hmakers/pov_hmaker.h"
#include "objects/bounding_slab.h"

namespace RAYPP {

void POV_HMAKER::Find_Infinites (vector<HANDLE<OBJECT> > &Orig,
  vector<SORT_ENTRY> &Infinite, vector<SORT_ENTRY> &Finite) const
  {
  SORT_ENTRY Entry;
  for (vector<HANDLE<OBJECT> >::iterator i=Orig.begin(); i<Orig.end(); ++i)
    {
    Entry = SORT_ENTRY(*i);
    if (Entry.Box->Infinite())
      Infinite.push_back (Entry);
    else
      Finite.push_back (Entry);
    }
  Orig.clear();
  Message_Stream << "Scene contains " << Infinite.size() <<
    " infinite and " << Finite.size() << " finite objects." << endl;
  }

void POV_HMAKER::Remerge (vector<SORT_ENTRY> &Infinite,
  vector<SORT_ENTRY> &Finite, vector<HANDLE<OBJECT> > &New) const
  {
  uint4 m;
  for (m=0; m<Infinite.size(); ++m) New.push_back (Infinite[m].Obj);
  Infinite.clear();
  for (m=0; m<  Finite.size(); ++m) New.push_back (  Finite[m].Obj);
  Finite.clear();
  }

struct xcomp
  {
  bool operator()(const SORT_ENTRY &a, const SORT_ENTRY &b)
    { return (a.Center->x < b.Center->x); }
  };
struct ycomp
  {
  bool operator()(const SORT_ENTRY &a, const SORT_ENTRY &b)
    { return (a.Center->y < b.Center->y); }
  };
struct zcomp
  {
  bool operator()(const SORT_ENTRY &a, const SORT_ENTRY &b)
    { return (a.Center->z < b.Center->z); }
  };

typedef enum {nosort,xsort,ysort,zsort} SORT_TYPE;

bool POV_HMAKER::Find_Split (vector<SORT_ENTRY> &Arr, uint4 Lo, uint4 Hi,
  int4 &Pos) const
  {
  uint4 Positions = Hi - Lo - 1;
  SORT_TYPE Sort_Type = nosort;
  float8 Minval, tmpval;
  Pos = -1;
  float8 *Left = new float8[Positions], *Right = new float8[Positions];
  AXISBOX LBox, RBox;
  uint4 m;

  sort (Arr.begin()+Lo, Arr.begin()+Hi, xcomp());

  for (m=0; m<Positions; ++m)
    {
    LBox.Include (*(Arr[Lo+m].Box));
    RBox.Include (*(Arr[Hi-m-1].Box));
    Left[m] = (m+1)*LBox.SurfaceArea ();
    Right[Positions-m-1] = (m+1)*RBox.SurfaceArea ();
    }

  LBox.Include (*(Arr[Hi-1].Box));
  Minval = (Hi-Lo-3)*LBox.SurfaceArea ();
// +1 or -1 or -2 ????

  for (m=0; m<Positions; ++m)
    {
    tmpval = Left[m] + Right[m];
    if (tmpval < Minval)
      {
      Minval = tmpval;
      Pos = m;
      Sort_Type = xsort;
      }
    }

  sort (Arr.begin()+Lo, Arr.begin()+Hi, ycomp());
  LBox.Reset(); RBox.Reset();
  for (m=0; m<Positions; ++m)
    {
    LBox.Include (*(Arr[Lo+m].Box));
    RBox.Include (*(Arr[Hi-m-1].Box));
    Left[m] = (m+1)*LBox.SurfaceArea ();
    Right[Positions-m-1] = (m+1)*RBox.SurfaceArea ();
    }
  for (m=0; m<Positions; ++m)
    {
    tmpval = Left[m] + Right[m];
    if (tmpval < Minval)
      {
      Minval = tmpval;
      Pos = m;
      Sort_Type = ysort;
      }
    }

  sort (Arr.begin()+Lo, Arr.begin()+Hi, zcomp());
  LBox.Reset(); RBox.Reset();
  for (m=0; m<Positions; ++m)
    {
    LBox.Include (*(Arr[Lo+m].Box));
    RBox.Include (*(Arr[Hi-m-1].Box));
    Left[m] = (m+1)*LBox.SurfaceArea ();
    Right[Positions-m-1] = (m+1)*RBox.SurfaceArea ();
    }
  for (m=0; m<Positions; ++m)
    {
    tmpval = Left[m] + Right[m];
    if (tmpval < Minval)
      {
      Minval = tmpval;
      Pos = m;
      Sort_Type = zsort;
      }
    }

  delete[] Left; delete[] Right;

  Pos += Lo;
  if (Sort_Type == nosort) return false;

  if (!(Sort_Type == zsort))
    {
    if (Sort_Type == xsort)
      sort (Arr.begin()+Lo, Arr.begin()+Hi, xcomp());
    else
      sort (Arr.begin()+Lo, Arr.begin()+Hi, ycomp());
    }
  return true;
  }

void POV_HMAKER::Split (vector<SORT_ENTRY> &In, uint4 Lo, uint4 Hi, 
  vector<SORT_ENTRY> &Out) const
  {
  uint4 m;
  int4 Where;
  uint4 NumObjects = Hi - Lo;
  if (NumObjects <= Bunching_Factor)
    {
    if (NumObjects == 1)
      {
      Out.push_back (In[Lo]);
      }
    else
      {
      BOUNDING_SLAB *bslab = Cnew BOUNDING_SLAB;
      for (m=Lo; m<Hi; ++m) bslab->Add (In[m].Obj);
      bslab->Init ();
      Out.push_back (SORT_ENTRY (HANDLE<OBJECT> (bslab)));
      }
    }
  else
    {
    if (Find_Split (In, Lo, Hi, Where))
      {
      Split (In, Lo, Where+1, Out);
      Split (In, Where+1, Hi, Out);
      }
    else
      {
      BOUNDING_SLAB *bslab = Cnew BOUNDING_SLAB;
      for (m=Lo; m<Hi; ++m) bslab->Add (In[m].Obj);
      bslab->Init ();
      Out.push_back (SORT_ENTRY (HANDLE<OBJECT> (bslab)));
      }
    }
  }

//virtual
void POV_HMAKER::Init ()
  {
  if (initialized) return;
  if ((Bunching_Factor < 2) || (Bunching_Factor > 20))
    error ("POV_HMAKER: Bunching_Factor out of range");
  initialized = true;
  }

//virtual
void POV_HMAKER::Build_Hierarchy (vector<HANDLE<OBJECT> > &Arr) const
  {
  ci();
  vector<SORT_ENTRY> Infinite, Finite, New_Fin;

  Find_Infinites (Arr, Infinite, Finite);
  while (Finite.size() > Bunching_Factor)
    {
    Split (Finite, 0, Finite.size(), New_Fin);
    Finite = New_Fin;
    New_Fin.clear ();
    }
  Remerge (Infinite, Finite, Arr);
  }

void POV_HMAKER::Set_Bunching_Factor (uint2 new_factor)
  {
  cni();
  Bunching_Factor = new_factor;
  }

} // namespace RAYPP
